@@ -83,6 +83,9 @@ META = 'meta.xml'
 THUMBNAILS = 'Thumbnails'
 output_encode = genshi.output.encode
 EtreeElement = lxml.etree.Element
+XML_INVALID_CHAR_EXPR = re.compile(
+    # from https://www.w3.org/TR/REC-xml/#charsets
+    '[\x00-\x08\x0b\x0c\x0e-\x1F\uD800-\uDFFF\uFFFE\uFFFF]')
 
 # A note regarding OpenDocument namespaces:
 #
@@ -245,6 +248,14 @@ def update_py_attrs(node, value):
                 (node.attrib[py_attrs_attr], value)
 
 
+def escape_xml_invalid_chars(value, repl=' '):
+    "Replace invalid characters for XML."
+    if isinstance(value, str):
+        return XML_INVALID_CHAR_EXPR.sub(repl, value)
+    else:
+        return value
+
+
 class Template(MarkupTemplate):
 
     def __init__(self, source, filepath=None, filename=None, loader=None,
@@ -408,9 +419,14 @@ class Template(MarkupTemplate):
             if directive in GENSHI_CLOSING_DIRECTIVE:
                 # map closing tags with their opening tag
                 if is_opening:
-                    opened_tags.append(statement)
+                    opened_tags.append((statement, directive))
                 else:
-                    closing_tags[id(opened_tags.pop())] = statement
+                    opening_statement, opening_directive = opened_tags.pop()
+                    assert directive == opening_directive, (
+                        "Wrong pairing tags between '%s' and '%s'"
+                        % (opening_statement.text.encode('utf-8'),
+                           statement.text.encode('utf-8')))
+                    closing_tags[id(opening_statement)] = statement
             # - we only need to return opening statements
             if is_opening:
                 r_statements.append((statement,
@@ -526,7 +542,8 @@ class Template(MarkupTemplate):
                 if (grand_parent is None
                         or grand_parent.tag != table_cell_tag
                         ) or len(grand_parent) != 1 or has_style(parent):
-                    r_node.attrib[py_replace] = expr
+                    r_node.attrib[py_replace] = (
+                        '__relatorio_escape_invalid_chars(%s)' % expr)
                     continue
 
                 cache_id = id(r_node)
@@ -789,6 +806,7 @@ class Template(MarkupTemplate):
             type_ = 'float'
         elif isinstance(val, str):
             type_ = 'string'
+            val = escape_xml_invalid_chars(val)
         elif isinstance(val, datetime.timedelta):
             type_ = 'time'
             val = 'P%sD%sS' % (val.days, val.seconds)
@@ -809,6 +827,7 @@ class Template(MarkupTemplate):
                                                     kwargs)
         kwargs['__relatorio_make_dimension'] = ImageDimension(self.namespaces)
         kwargs['__relatorio_guess_type'] = self._guess_type
+        kwargs['__relatorio_escape_invalid_chars'] = escape_xml_invalid_chars
 
         counter = ColumnCounter()
         kwargs['__relatorio_reset_col_count'] = counter.reset
@@ -873,7 +892,8 @@ def get_zip_file(source):
 def fod2od(source):
     "Convert Flat OpenDocument to OpenDocument"
     odt_io = BytesIO()
-    odt_zip = zipfile.ZipFile(odt_io, mode='w')
+    odt_zip = zipfile.ZipFile(
+        odt_io, mode='w', compression=zipfile.ZIP_DEFLATED)
     fodt_tree = lxml.etree.parse(source)
     fodt_root = fodt_tree.getroot()
     office_ns = fodt_root.nsmap['office']
@@ -888,6 +908,8 @@ def fod2od(source):
         '{%s}body' % office_ns: ['content'],
         }
     mimetype = fodt_root.attrib['{%s}mimetype' % office_ns]
+    # mimetype should be written first to let detection through 'magic number'
+    odt_zip.writestr('mimetype', mimetype, zipfile.ZIP_STORED)
     documents = {}
     images = []
     for child in fodt_root:
@@ -916,7 +938,6 @@ def fod2od(source):
         odt_zip.writestr(fname, data)
         manifest.add_file_entry(fname, mime_type)
     odt_zip.writestr(MANIFEST, str(manifest))
-    odt_zip.writestr('mimetype', mimetype)
     odt_zip.close()
     return odt_io
 
@@ -1026,7 +1047,8 @@ class OOSerializer:
         self.manifest = Manifest(self.inzip.read(MANIFEST))
         self.meta = Meta(self.inzip.read(META))
         self.new_oo = BytesIO()
-        self.outzip = zipfile.ZipFile(self.new_oo, 'w')
+        self.outzip = zipfile.ZipFile(
+            self.new_oo, mode='w', compression=zipfile.ZIP_DEFLATED)
         self.xml_serializer = genshi.output.XMLSerializer()
 
     def __call__(self, stream):
